/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.android.tools.idea.resourceExplorer.importer

import com.android.resources.ResourceFolderType
import com.android.tools.idea.res.ResourceRepositoryManager
import com.android.tools.idea.resourceExplorer.model.DesignAssetSet
import com.intellij.openapi.Disposable
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.vfs.*
import com.intellij.util.io.DigestUtil
import org.jetbrains.android.facet.AndroidFacet
import org.jetbrains.android.util.AndroidResourceUtil

/**
 * Manager to keep track of the differences between the facet's modules' resources and
 * provided [VirtualFile]
 */
class SynchronizationManager(facet: AndroidFacet)
  : VirtualFileListener,
    Disposable {

  private var resourcesRepository = ResourceRepositoryManager.getModuleResources(facet)
  private var hashToFile: MutableMap<String, VirtualFile> = mapAllModuleResources(resourcesRepository.resourceDirs)
  private val listeners: MutableList<SynchronizationListener> = mutableListOf()

  init {
    VirtualFileManager.getInstance().addVirtualFileListener(this)
    Disposer.register(this, resourcesRepository)
  }

  /**
   * Map the hash of the resources to the virtual file
   */
  private fun mapAllModuleResources(resourcesDirectories: MutableSet<VirtualFile>)
      : MutableMap<String, VirtualFile> {
    return AndroidResourceUtil
        .getResourceSubdirs(ResourceFolderType.DRAWABLE, resourcesDirectories)
        .flatMap { it.allChildren() }
        .associate { sha1(it) to it }.toMutableMap()
  }

  /**
   * Check if the files in [assetSet] are synchronized with the resources of the [AndroidFacet]
   */
  fun getSynchronizationStatus(assetSet: DesignAssetSet): SynchronizationStatus {
    return when (assetSet.designAssets.count { hashToFile.containsKey(sha1(it.file)) }) {
      assetSet.designAssets.size -> SynchronizationStatus.SYNCED
      0 -> SynchronizationStatus.NOT_SYNCED
      else -> SynchronizationStatus.PARTIALLY_SYNCED
    }
  }

  /**
   * Check if the file is in any of the resources directories of the facet
   */
  private fun isModuleResourceFile(virtualFile: VirtualFile) =
      !virtualFile.isDirectory && resourcesRepository.resourceDirs.any { VfsUtilCore.isAncestor(it, virtualFile, false) }

  private fun notifyResourceCreated(file: VirtualFile) {
    listeners.forEach { it.resourceAdded(file) }
  }

  private fun notifyResourceRemoved(file: VirtualFile) {
    listeners.forEach { it.resourceRemoved(file) }
  }

  override fun fileCreated(event: VirtualFileEvent) {
    if (isModuleResourceFile(event.file)) {
      hashToFile.put(sha1(event.file), event.file)
      notifyResourceCreated(event.file)
    }
  }

  override fun fileDeleted(event: VirtualFileEvent) {
    if (isModuleResourceFile(event.file)) {
      hashToFile.entries
          .firstOrNull { it.value == event.file }
          ?.key
          ?.let(hashToFile::remove)
      notifyResourceRemoved(event.file)
    }
  }

  fun addListener(listener: SynchronizationListener) {
    listeners.add(listener)
  }

  fun removeListener(listener: SynchronizationListener) {
    listeners.remove(listener)
  }

  override fun dispose() {
    VirtualFileManager.getInstance().removeVirtualFileListener(this)
  }
}

/**
 * Enum that represent the synchronisation status of a [DesignAssetSet]
 */
enum class SynchronizationStatus {

  /**
   * None of the files of the [DesignAssetSet] is present in the facet's module
   */
  NOT_SYNCED,

  /**
   * All the files of the [DesignAssetSet] are present in the facet's module
   */
  SYNCED,

  /**
   * Some of the files of the [DesignAssetSet] are present in the facet's module
   */
  PARTIALLY_SYNCED
}

/**
 * Listener to notify when a resources has been added or removed from the module associated
 * with the SynchronizationManager
 */
interface SynchronizationListener {
  fun resourceAdded(file: VirtualFile)
  fun resourceRemoved(file: VirtualFile)
}

private fun VirtualFile.allChildren(): List<VirtualFile> =
    if (!isDirectory) listOf(this) else children.flatMap { it.allChildren() }

private fun sha1(virtualFile: VirtualFile): String {
  return DigestUtil.sha1().digest(virtualFile.contentsToByteArray()).toString(Charsets.UTF_8)
}
